@@ -1,7 +1,7 @@
 """
 Components for emulating an x86/x64 architecture.
 """
-
+from .instruction import x86_64Instruction
 from ..cpu_context import ProcessorContext
 from .registers import x86_64_Registers
 from .opcodes import OPCODES
@@ -17,8 +17,8 @@ RBP_OFFSET = 0x400
 class x86_64ProcessorContext(ProcessorContext):
     """Processor context for x86/x64 architecture"""
 
-    ARCH_NAME = "metapc"
     OPCODES = OPCODES.copy()
+    _instruction_class = x86_64Instruction
 
     def __init__(self, emulator):
         super(x86_64ProcessorContext, self).__init__(
@@ -26,6 +26,7 @@ class x86_64ProcessorContext(ProcessorContext):
             registers=x86_64_Registers(),
             instruction_pointer="rip",
             stack_pointer="rsp",
+            return_register="rax",
         )
         # Set up the stack before we go.
         self.registers.rsp = STACK_BASE - RSP_OFFSET
